(*
  AtariHD - read/write Atari disks on the PC

  Copyright (c) 1992-2003 by Matthias Reichl <hias@horus.com>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)

IMPLEMENTATION MODULE AtariHD;

IMPORT SYSTEM,Lib,IO,FIO;

VAR LastStatus : SHORTCARD;

VAR drive:SHORTCARD;
    Invert:BOOLEAN;
    drive_type:SHORTCARD; (* type of drive, as listed in CMOS *)

VAR DebugLevel : DebugLevelType;
    Interactive : BOOLEAN;

VAR FormatInterleaveAtari:BOOLEAN;


VAR SectorsPerTrack,TracksPerSide,Sides,SectorLength:CARDINAL;

TYPE SectorMapping = ( Mapping_Track, (* S/T/H, change tracks first, then heads *)
                       Mapping_XF551, (* S/T/H, reverse side 2 *)
                       Mapping_Head   (* S/H/T, change heads first, then tracks *)
                     );
TYPE SecIOType = ( SecIO_Read,
                   SecIO_Write,
                   SecIO_Verify
                 );

TYPE OperationType = ( Op_Read,
                       Op_Write,
                       Op_Verify,
                       Op_Format
                     );

TYPE ARIResultType = ( ARI_Abort,
                       ARI_Retry,
                       ARI_Ignore
                     );

VAR CurrentMapping : SectorMapping;

PROCEDURE SetDebugLevel(level: DebugLevelType);
BEGIN
    DebugLevel := level;
END SetDebugLevel;

PROCEDURE SetInteractive(interactive: BOOLEAN);
BEGIN
    Interactive := interactive;
END SetInteractive;

PROCEDURE GetBytesPerSector() : CARDINAL;
BEGIN
    RETURN SectorLength;
END GetBytesPerSector;

PROCEDURE GetSectorsPerTrack() : CARDINAL;
BEGIN
    RETURN SectorsPerTrack;
END GetSectorsPerTrack;

PROCEDURE GetTracksPerSide() : CARDINAL;
BEGIN
    RETURN TracksPerSide;
END GetTracksPerSide;

PROCEDURE GetSides() : CARDINAL;
BEGIN
    RETURN Sides;
END GetSides;

VAR disk_param_table[0:78H]:ADDRESS;
    old_table:ADDRESS;
TYPE a11=ARRAY[0..10] OF BYTE;
TYPE a11p=POINTER TO a11;
VAR new_table:a11;

CONST MaxRetry=5;

TYPE Array18=ARRAY[0..17] OF SHORTCARD;
     Array26=ARRAY[0..25] OF SHORTCARD;
     Array36=ARRAY[0..35] OF SHORTCARD;

CONST InterleaveDD=Array18(1,12,4,15,7,18,10,2,13,5,16,8,11,3,14,6,17,9);
CONST InterleaveSD=Array18(1,3,5,7,9,11,13,15,17,2,4,6,8,10,12,14,16,18);
(*
      InterleaveHD=Array36(1,15,29,7,21,35,13,27,5,19,33,11,25,3,17,31,9,23,
                           26,4,18,32,10,24,2,16,30,8,22,36,14,28,6,20,34,12);
*)

      InterleaveED=Array26(
                   1, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 23, 25,
                   2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 26
                   );

VAR InterleaveUser:Array36;
    ReadInterleave:Array36;


PROCEDURE SaveOldTable;
BEGIN
    old_table:=disk_param_table
END SaveOldTable;

PROCEDURE RestoreOldTable;
BEGIN
    disk_param_table:=old_table
END RestoreOldTable;

PROCEDURE SetNewTable;
VAR p:a11p;
BEGIN
    p:=a11p(old_table);
    new_table:=p^;
    new_table[3]:=1;    (* 256 bytes per sector *)
    new_table[4]:=12H;  (* 18 sectors per track *)

    new_table[5]:=10;  (* gap 3 length for write *)
    new_table[7]:=20;  (* gap 3 length for format *)

(*
    new_table[5]:=0CH;  (* gap 3 length for write *)
    new_table[7]:=1CH;  (* gap 3 length for format *)
*)

    new_table[8]:=0FFH;
(*

    new_table:=a11(0DFH,02H,025H,
                   01H,
                   3FH,
(*
                   2AH,080H,30H,0F6H,014H,08H);
                   1BH,0FFH,54H,0F6H,0FH,08H);
                   1FH,0FFH,24H,0FFH,0FH,08H);
                   05H,0FFH,10H,0FFH,0FH,08H);
*)
                   0CH,0FFH,1CH,0FFH,0FH,08H); (* 29,24*)
*)
    disk_param_table:=ADR(new_table);
END SetNewTable;

PROCEDURE InitController;
 VAR r:SYSTEM.Registers;
BEGIN
    r.AH:=0;
    r.DL:=drive;
    Lib.Intr(r,13H);
END InitController;

(*
PROCEDURE GetLastStatus() : SHORTCARD;
 VAR r:SYSTEM.Registers;
BEGIN
    r.AH:=1;
    r.DL:=drive;
    Lib.Intr(r,13H);
    RETURN r.AH;
END GetLastStatus;
*)

PROCEDURE SecToSTH(sec:CARDINAL; VAR s,t,h:SHORTCARD);
BEGIN
    DEC(sec);

    CASE CurrentMapping OF
    | Mapping_Track, Mapping_XF551:
        s:=SHORTCARD((sec MOD SectorsPerTrack)+1);
        t:=SHORTCARD((sec DIV SectorsPerTrack) MOD TracksPerSide);
        h:=SHORTCARD((sec DIV (SectorsPerTrack * TracksPerSide)) MOD Sides);
        IF (CurrentMapping = Mapping_XF551) AND (h=1) THEN
            t := SHORTCARD(TracksPerSide) - t - 1;
            s := SHORTCARD(SectorsPerTrack) + 1 - s ;
        END;
    | Mapping_Head:
        s:=SHORTCARD((sec MOD SectorsPerTrack)+1);
        h:=SHORTCARD((sec DIV SectorsPerTrack) MOD Sides);
        t:=SHORTCARD((sec DIV (SectorsPerTrack*Sides)) MOD TracksPerSide);
    END;
END SecToSTH;

PROCEDURE TrackToTH(track:CARDINAL; VAR t,h:SHORTCARD);
BEGIN
    CASE CurrentMapping OF
    | Mapping_Track, Mapping_XF551:
        t:=SHORTCARD(track MOD TracksPerSide);
        h:=SHORTCARD((track DIV TracksPerSide) MOD Sides);
        IF (CurrentMapping = Mapping_XF551) AND (h=1) THEN
            t := SHORTCARD(TracksPerSide) - t - 1;
        END;
    | Mapping_Head:
        h:=SHORTCARD(track MOD Sides);
        t:=SHORTCARD((track DIV Sides) MOD TracksPerSide);
    END;
END TrackToTH;

PROCEDURE STHToSec(sector,track,head : SHORTCARD) : CARDINAL;
VAR atariSec : CARDINAL;
BEGIN
    atariSec := 0;
    CASE CurrentMapping OF
    | Mapping_XF551, Mapping_Track:
        IF (CurrentMapping = Mapping_XF551) AND (head = 1) THEN
            track := SHORTCARD(TracksPerSide) - track - 1;
            sector := SHORTCARD(SectorsPerTrack) + 1 - sector;
        END;
        atariSec :=
             CARDINAL(sector)
           + CARDINAL(track) * SectorsPerTrack
           + CARDINAL(head) * TracksPerSide*SectorsPerTrack;
        RETURN atariSec;
    | Mapping_Head:
        atariSec :=
             CARDINAL(sector)
           + CARDINAL(head) * (SectorsPerTrack*(Sides-1))
           + CARDINAL(track) * (SectorsPerTrack*Sides);
        RETURN atariSec;
    END;
    HALT;
END STHToSec;

PROCEDURE AbortRetryIgnore(
                            operation: OperationType;
                            sector, track, head: SHORTCARD
                          ) : ARIResultType;
VAR c: CHAR;
LABEL keyLoop;
BEGIN
    IO.WrLn;
    IO.WrStr("Error ");
    CASE operation OF
    | Op_Read: IO.WrStr("reading sector");
    | Op_Write: IO.WrStr("writing sector");
    | Op_Verify: IO.WrStr("verifying sector");
    | Op_Format: IO.WrStr("formatting track");
    END;
    IO.WrStr(" (status=");
    IF (LastStatus<16) THEN IO.WrStr("0"); END;
    IO.WrShtHex(LastStatus,0); IO.WrStr(", ");
    CASE LastStatus OF
    | 00H:	IO.WrStr("successful completion");
    | 01H:	IO.WrStr("invalid function in AH or invalid parameter");
    | 02H:	IO.WrStr("address mark not found");
    | 03H:	IO.WrStr("disk write-protected");
    | 04H:	IO.WrStr("sector not found/read error");
    | 05H:	IO.WrStr("reset failed");
    | 06H:	IO.WrStr("disk changed");
    | 07H:	IO.WrStr("drive parameter activity failed");
    | 08H:	IO.WrStr("DMA overrun");
    | 09H:	IO.WrStr("data boundary error");
    | 0AH:	IO.WrStr("bad sector detected");
    | 0BH:	IO.WrStr("bad track detected");
    | 0CH:	IO.WrStr("unsupported track or invalid media");
    | 0DH:	IO.WrStr("invalid number of sectors on format");
    | 0EH:	IO.WrStr("control data address mark detected");
    | 0FH:	IO.WrStr("DMA arbitration level out of range");
    | 10H:	IO.WrStr("uncorrectable CRC or ECC error");
    | 11H:	IO.WrStr("data ECC corrected");
    | 20H:	IO.WrStr("controller failure");
    | 31H:	IO.WrStr("no media in drive");
    | 32H:	IO.WrStr("incorrect drive type stored in CMOS");
    | 40H:	IO.WrStr("seek failed");
    | 80H:	IO.WrStr("timeout");
    | 0AAH:	IO.WrStr("drive not ready");
    | 0B0H:	IO.WrStr("volume not locked in drive");
    | 0B1H:	IO.WrStr("volume locked in drive");
    | 0B2H:	IO.WrStr("volume not removable");
    | 0B3H:	IO.WrStr("volume in use");
    | 0B4H:	IO.WrStr("lock count exceeded");
    | 0B5H:	IO.WrStr("valid eject request failed");
    | 0B6H:	IO.WrStr("volume present but read protected");
    | 0BBH:	IO.WrStr("undefined error");
    | 0CCH:	IO.WrStr("write fault");
    | 0E0H:	IO.WrStr("status register error");
    | 0FFH:	IO.WrStr("sense operation failed");
    ELSE
       IO.WrStr("unknown error");
    END;
    IO.WrChar(')'); IO.WrLn;
    IO.WrStr("on drive "); IO.WrChar('A'+CHAR(drive));
    IO.WrChar(' ');
    IF (operation # Op_Format) THEN
        IO.WrStr("sector "); IO.WrShtCard(sector,0);
        IO.WrChar('/');
    END;
    IO.WrStr("track "); IO.WrShtCard(track,0);
    IO.WrStr("/head "); IO.WrShtCard(head,0);
    IF (operation # Op_Format) THEN
        IO.WrStr(" (Atari sector ");
        IO.WrCard(STHToSec(sector,track,head),0);
        IO.WrChar(')');
    END;
    IO.WrLn;
    IO.WrStr("(A)bort (R)etry (I)gnore? ");

keyLoop:
    c := IO.RdKey();
    IF (c=CHAR(0)) THEN (* ignore function keys *)
       c:=IO.RdKey();
       IO.WrChar(CHAR(7));
       GOTO keyLoop;
    END;
    c := CAP(c);
    IF ( (c#'A') AND (c#'R') AND (c#'I') ) THEN
       IO.WrChar(CHAR(7));
       GOTO keyLoop;
    END;
    IO.WrChar(c); IO.WrLn;
    CASE c OF
    | 'A': RETURN ARI_Abort;
    | 'R': RETURN ARI_Retry;
    | 'I': RETURN ARI_Ignore;
    END;
    RETURN ARI_Abort;
END AbortRetryIgnore;

PROCEDURE IntSectorIO( operation: SecIOType;
                       Sector,Track,Head,Count:SHORTCARD;
                       VAR buf:ARRAY OF BYTE);
VAR r:SYSTEM.Registers;
    retry:CARDINAL;
LABEL again;
BEGIN
again:
    retry:=0;
    REPEAT
        CASE operation OF
        | SecIO_Read:
           r.AH:=2;
        | SecIO_Write:
           r.AH:=3;
        | SecIO_Verify:
           r.AH:=4;
        ELSE
           ok := FALSE;
           LastStatus := 254;
           RETURN;
        END;
        r.DL:=drive;
        r.DH:=Head;
        r.CH:=Track;
        r.CL:=Sector;
        r.AL:=Count;
        r.ES:=SYSTEM.Seg(buf);
        r.BX:=SYSTEM.Ofs(buf);
        Lib.Intr(r,13H);
        ok := (r.Flags * {SYSTEM.CarryFlag}={});
        IF ~ok THEN
            LastStatus := r.AH;
            InitController;
        END;
        INC(retry);
    UNTIL ok OR (retry=MaxRetry);
    IF (~ok) AND Interactive THEN
       CASE AbortRetryIgnore(OperationType(operation), Sector, Track, Head) OF
       | ARI_Ignore: ok := TRUE;
       | ARI_Retry: GOTO again;
       END;
    END;
END IntSectorIO;


PROCEDURE FormatTrack(track:CARDINAL);
VAR trkinfo:ARRAY[0..100] OF RECORD
                               track,side,secno,byte:SHORTCARD
                            END;
    i:CARDINAL;
    r:SYSTEM.Registers;
    retry:CARDINAL;

    s,t,h:SHORTCARD;

    spt : CARDINAL;

LABEL again;

BEGIN
    TrackToTH(track,t,h);
    FOR i:=0 TO SectorsPerTrack-1 DO
        trkinfo[i].track:=t;
        trkinfo[i].side:=h;
        s:=SHORTCARD(i+1);
        IF FormatInterleaveAtari THEN
            s:=InterleaveUser[i]
        END;
        IF (CurrentMapping = Mapping_XF551) AND (h=1) THEN
           s:=SHORTCARD(SectorsPerTrack)+1-s;
        END;
        trkinfo[i].secno:=s;

        IF SectorLength=128 THEN
            trkinfo[i].byte:=0;
        ELSIF SectorLength=256 THEN
            trkinfo[i].byte:=1;
        ELSE
            trkinfo[i].byte:=2;
        END;
    END;

again:
    retry:=0;
    REPEAT
        r.AH:=5;
        r.DL:=drive;
        r.DH:=h;
        r.CH:=t;
        r.CL:=0;
        r.AL:=SHORTCARD(spt);
        r.ES:=SYSTEM.Seg(trkinfo);
        r.BX:=SYSTEM.Ofs(trkinfo);
        Lib.Intr(r,13H);
        ok := (r.Flags * {SYSTEM.CarryFlag}={});
        IF ~ok THEN
            LastStatus := r.AH;
            InitController;
        END;
        INC(retry);
    UNTIL ok OR (retry=MaxRetry);
    IF (~ok) AND Interactive THEN
       CASE AbortRetryIgnore(Op_Format, 0, t, h) OF
       | ARI_Ignore: ok := TRUE;
       | ARI_Retry: GOTO again;
       END;
    END;
END FormatTrack;


PROCEDURE FormatDisk;
VAR track:CARDINAL;
    max : CARDINAL;
BEGIN
    track:=0;
    max := TracksPerSide*Sides;
    REPEAT
         FormatTrack(track);
         INC(track);
    UNTIL (track=max) OR (ok=FALSE)
END FormatDisk;

PROCEDURE WriteSector(sec:CARDINAL; buf:ARRAY OF BYTE);
VAR r:SYSTEM.Registers;
    i:CARDINAL;
    s,t,h:SHORTCARD;
    retry:CARDINAL;
BEGIN
    IF Invert THEN
        FOR i:=0 TO SectorLength-1 DO
            buf[i]:=BYTE(255-SHORTCARD(buf[i]))
        END;
    END;

    SecToSTH(sec,s,t,h);

    IntSectorIO(SecIO_Write,s,t,h,1,buf);

END WriteSector;

PROCEDURE ReadSector(sec:CARDINAL; VAR buf:ARRAY OF BYTE);
VAR r:SYSTEM.Registers;
    i:CARDINAL;
    s,t,h:SHORTCARD;
    retry:CARDINAL;
BEGIN
    retry:=0;

    SecToSTH(sec,s,t,h);
    IntSectorIO(SecIO_Read,s,t,h,1,buf);

    IF (ok & Invert) THEN
        FOR i:=0 TO SectorLength-1 DO
            buf[i]:=BYTE(255-SHORTCARD(buf[i]))
        END;
    END;
END ReadSector;

PROCEDURE VerifySector(sec:CARDINAL; buf:ARRAY OF BYTE);
VAR r:SYSTEM.Registers;
    i:CARDINAL;
    s,t,h:SHORTCARD;
    retry:CARDINAL;
BEGIN
    IF Invert THEN
        FOR i:=0 TO SectorLength-1 DO
            buf[i]:=BYTE(255-SHORTCARD(buf[i]))
        END;
    END;

    SecToSTH(sec,s,t,h);

    IntSectorIO(SecIO_Verify,s,t,h,1,buf);
END VerifySector;


PROCEDURE ReadTrack(track:CARDINAL; VAR buf:ARRAY OF BYTE);
VAR i:CARDINAL;
    s,sDisk,t,h:SHORTCARD;
BEGIN
    TrackToTH(track,t,h);
    i:=0;
    ok:=TRUE;
    WHILE (i<SectorsPerTrack) AND (ok) DO
        s:=ReadInterleave[i];

        IF (CurrentMapping = Mapping_XF551) AND (h=1) THEN
           sDisk:=SHORTCARD(SectorsPerTrack)+1-s;
        ELSE
           sDisk:=s;
        END;

        IntSectorIO(SecIO_Read,sDisk,t,h,1,buf[CARDINAL(s-1)*SectorLength]);
        INC(i);
    END;
    IF (ok) AND Invert THEN
       FOR i:=0 TO SectorsPerTrack*SectorLength-1 DO
           buf[i]:=BYTE(255-SHORTCARD(buf[i]));
       END;
    END;
END ReadTrack;

PROCEDURE WriteTrack(track:CARDINAL; buf:ARRAY OF BYTE);
VAR i:CARDINAL;
    s,sDisk, t,h:SHORTCARD;
    ok2:BOOLEAN;
BEGIN
    IF Invert THEN
       FOR i:=0 TO SectorsPerTrack*SectorLength-1 DO
           buf[i]:=BYTE(255-SHORTCARD(buf[i]));
       END;
    END;
    TrackToTH(track, t, h);
    i:=0;
    ok:=TRUE;
    WHILE (i<SectorsPerTrack) AND (ok) DO
        s:=ReadInterleave[i];
        IF (CurrentMapping = Mapping_XF551) AND (h=1) THEN
           sDisk:=SHORTCARD(SectorsPerTrack)+1-s;
        ELSE
           sDisk:=s;
        END;
        IntSectorIO(SecIO_Write,sDisk,t,h,1,buf[CARDINAL(s-1)*SectorLength]);
        INC(i);
    END;
END WriteTrack;

PROCEDURE VerifyTrack(track:CARDINAL; buf:ARRAY OF BYTE);
VAR i:CARDINAL;
    s,sDisk, t,h:SHORTCARD;
    ok2:BOOLEAN;
BEGIN
    IF Invert THEN
       FOR i:=0 TO SectorsPerTrack*SectorLength-1 DO
           buf[i]:=BYTE(255-SHORTCARD(buf[i]));
       END;
    END;
    TrackToTH(track, t, h);
    i:=0;
    ok:=TRUE;
    WHILE (i<SectorsPerTrack) AND (ok) DO
        s:=ReadInterleave[i];
        IF (CurrentMapping = Mapping_XF551) AND (h=1) THEN
           sDisk:=SHORTCARD(SectorsPerTrack)+1-s;
        ELSE
           sDisk:=s;
        END;
        IntSectorIO(SecIO_Verify,sDisk,t,h,1,buf[CARDINAL(s-1)*SectorLength]);
        INC(i);
    END;
END VerifyTrack;

PROCEDURE Fix360;
(* Sets correct format-type for 360k-disk (in 360k/1.2m drive) *)
VAR r:SYSTEM.Registers;
BEGIN
    r.AH:=17H;
    r.DL:=drive;
    CASE drive_type OF
      1: r.AL:=1; Lib.Intr(r,13H);
    | 2: r.AL:=2; Lib.Intr(r,13H);
    END;
    IF (DebugLevel >= DebugLevel_Normal) THEN
        CASE drive_type OF
          1: IO.WrStr("Preparing controller for a 360k disk in a 360k drive");
             IO.WrLn;
        | 2: IO.WrStr("Preparing controller for a 360k disk in a 1.2M drive");
             IO.WrLn;
        END;
    END;
END Fix360;


PROCEDURE Fix720;
(* Sets correct format-type for 720k-disk (in 1.44m drive) *)
VAR r:SYSTEM.Registers;
BEGIN
    IF (drive_type=3) OR (drive_type=4) THEN
        r.AH:=17H;
        r.DL:=drive;
        r.AL:=4;
        Lib.Intr(r,13H);
        IF (DebugLevel >= DebugLevel_Normal) THEN
            IO.WrStr("Preparing controller for a 720k disk");
            IO.WrLn;
        END;
    END;
END Fix720;

PROCEDURE SetStandardGap;
BEGIN
    new_table[5]:=10;  (* gap 3 length for write *)
    new_table[7]:=20;  (* gap 3 length for format *)
END SetStandardGap;

PROCEDURE SetHDGap;
BEGIN
    new_table[5]:=6;  (* gap 3 length for write *)
    new_table[7]:=10;  (* gap 3 length for format *)
END SetHDGap;

PROCEDURE SetSDGap;
BEGIN
    new_table[5]:=10;  (* gap 3 length for write *)
    new_table[7]:=17;  (* gap 3 length for format *)
END SetSDGap;

PROCEDURE SetEDGap;
BEGIN
    new_table[5]:=16;  (* gap 3 length for write *)
    new_table[7]:=32;  (* gap 3 length for format *)
END SetEDGap;

PROCEDURE InitSD;
BEGIN
    SectorLength:=128;
    SectorsPerTrack:=18;
    TracksPerSide:=40;
    Sides:=1;
    new_table[3]:=0;
    new_table[4]:=BYTE(SectorsPerTrack+1);
    new_table[6]:=128;
    SetSDGap;
    Invert:=TRUE;
    InitController;
    CurrentMapping := Mapping_Track;
    SetInterleaveSD;
    Fix360;
END InitSD;

PROCEDURE InitED();
BEGIN
    SectorLength:=128;
    SectorsPerTrack:=26;
    TracksPerSide:=40;
    Sides:=1;
    new_table[3]:=0;
    new_table[4]:=BYTE(SectorsPerTrack);
    new_table[6]:=128;
    SetEDGap;
    Invert:=TRUE;
    InitController;
    CurrentMapping := Mapping_Track;
    SetInterleaveED;
    Fix360;
END InitED;

PROCEDURE InitDD;
BEGIN
    IF (DebugLevel >= DebugLevel_Normal) THEN
       IO.WrStr('Perparing for 180k/5.25" format'); IO.WrLn;
    END;
    SectorLength:=256;
    SectorsPerTrack:=18;
    TracksPerSide:=40;
    Sides:=1;
    new_table[3]:=1;
    new_table[4]:=BYTE(SectorsPerTrack);
    new_table[6]:=0FFH;
    SetStandardGap;
    Invert:=TRUE;
    InitController;
    CurrentMapping := Mapping_Track;
    SetInterleaveAtari;
    Fix360;
END InitDD;

PROCEDURE InitDSDD;
BEGIN
    IF (DebugLevel >= DebugLevel_Normal) THEN
       IO.WrStr('Perparing for 360k/5.25" format'); IO.WrLn;
    END;
    SectorLength:=256;
    SectorsPerTrack:=18;
    TracksPerSide:=40;
    Sides:=2;
    CurrentMapping:=Mapping_Track;
    new_table[3]:=1;
    new_table[4]:=BYTE(SectorsPerTrack);
    new_table[6]:=0FFH;
    SetStandardGap;
    Invert:=TRUE;
    InitController;
    CurrentMapping := Mapping_Track;
    SetInterleaveAtari;
    Fix360;
END InitDSDD;

PROCEDURE Init720_35;
BEGIN
    IF (DebugLevel >= DebugLevel_Normal) THEN
       IO.WrStr('Perparing for 720k/3.5" format'); IO.WrLn;
    END;
    SectorLength:=256;
    SectorsPerTrack:=18;
    TracksPerSide:=80;
    Sides:=2;
    new_table[3]:=1;
    new_table[4]:=BYTE(SectorsPerTrack);
    new_table[6]:=0FFH;
    SetStandardGap;
    Invert:=TRUE;
    CurrentMapping := Mapping_Track;
    InitController;
    SetInterleaveAtari;
    Fix720;
END Init720_35;

PROCEDURE Init1440_35;
BEGIN
    IF (DebugLevel >= DebugLevel_Normal) THEN
       IO.WrStr('Perparing for 1.44M/3.5" format'); IO.WrLn;
    END;
    SectorLength:=256;
    SectorsPerTrack:=36;
    TracksPerSide:=80;
    Sides:=2;
    new_table[3]:=1;
    new_table[4]:=BYTE(SectorsPerTrack);
    new_table[6]:=0FFH;
    SetHDGap;
    Invert:=TRUE;
    CurrentMapping := Mapping_Track;
    InitController;
    SetInterleaveHD;
END Init1440_35;

PROCEDURE InitPCDD;
BEGIN
    SectorLength:=512;
    SectorsPerTrack:=9;
    Sides:=2;
    CurrentMapping:=Mapping_Head;
    new_table[3]:=2;
    new_table[4]:=BYTE(SectorsPerTrack);
    new_table[6]:=0FFH;
    SetStandardGap;
    Invert:=FALSE;
    CurrentMapping := Mapping_Track;
    InitController;
    SetInterleave1;
END InitPCDD;

PROCEDURE CheckDriveCompatibility(type : FloppyType) : BOOLEAN;
BEGIN
    CASE type OF
    | FloppySD, FloppyED, FloppyDD, FloppyDSDD, FloppyDSDD_XF551, FloppyDSDD_HeadFirst :
       IF (drive_type = 1) OR (drive_type = 2) THEN
           RETURN TRUE;
       ELSE
           RETURN FALSE;
       END;
    | Floppy720_35, Floppy720_35_XF551, Floppy720_35_HeadFirst :
       IF (drive_type = 3) OR (drive_type = 4) THEN
          RETURN TRUE;
       ELSE
          RETURN FALSE;
       END;
    | Floppy1440_35, Floppy1440_35_HeadFirst :
       IF (drive_type = 4) THEN
          RETURN TRUE;
       ELSE
          RETURN FALSE;
       END;
    ELSE
       IO.WrStr("unsupported floppy type!"); IO.WrLn;
       RETURN FALSE;
    END;
END CheckDriveCompatibility;


PROCEDURE InitFloppy(type : FloppyType);
BEGIN
    CASE type OF
    | FloppyDD :
        InitDD;
        CurrentMapping := Mapping_Track;
    | FloppyDSDD_XF551 :
        InitDSDD;
        CurrentMapping := Mapping_XF551;
    | Floppy720_35_XF551 :
        Init720_35;
        CurrentMapping := Mapping_XF551;

    | FloppyDSDD :
        InitDSDD;
        CurrentMapping := Mapping_Track;
    | Floppy720_35 :
        Init720_35;
        CurrentMapping := Mapping_Track;
    | Floppy1440_35 :
        Init1440_35;
        CurrentMapping := Mapping_Track;

    | FloppyDSDD_HeadFirst :
        InitDSDD;
        CurrentMapping := Mapping_Head;
    | Floppy720_35_HeadFirst :
        Init720_35;
        CurrentMapping := Mapping_Head;
    | Floppy1440_35_HeadFirst :
        Init1440_35;
        CurrentMapping := Mapping_Head;

    | FloppyPCDD :
        InitPCDD;
        CurrentMapping := Mapping_Head;
    | FloppySD :
        InitSD;
        CurrentMapping := Mapping_Track;
    | FloppyED :
        InitED();
        CurrentMapping := Mapping_Track;
    ELSE
       IO.WrStr("unsupported floppy type!"); IO.WrLn;
       HALT;
    END;
END InitFloppy;




PROCEDURE CheckDriveType;
VAR r:SYSTEM.Registers;
    p:a11p;
BEGIN
 r.AH:=08H;
 r.DL:=drive;
 Lib.Intr(r,13H);
 drive_type:=SHORTCARD(BITSET(r.BL)*BITSET(15));
 p:=a11p([r.ES:r.DI]);
 IF (DebugLevel >= DebugLevel_Normal) THEN
    IO.WrStr("Drive ");
    IO.WrChar('A'+CHAR(drive));
    IO.WrStr(" is a ");
    CASE drive_type OF
    | 1: IO.WrStr("360k");
    | 2: IO.WrStr("1.2M");
    | 3: IO.WrStr("720k");
    | 4: IO.WrStr("1.44M");
    | 6: IO.WrStr("2.88M");
    ELSE
       IO.WrStr("unknown (type="); IO.WrShtCard(drive_type,0);
       IO.WrStr(")");
    END;
    IO.WrStr(" drive"); IO.WrLn;
 END;
END CheckDriveType;

PROCEDURE SetDriveA;
BEGIN
 IF (DebugLevel >= DebugLevel_Normal) THEN
    IO.WrStr("Initializing drive A"); IO.WrLn;
 END;
 drive:=0;
 InitController;
 CheckDriveType;
END SetDriveA;

PROCEDURE SetDriveB;
BEGIN
 IF (DebugLevel >= DebugLevel_Normal) THEN
    IO.WrStr("Initializing drive B"); IO.WrLn;
 END;
 drive:=1;
 InitController;
 CheckDriveType;
END SetDriveB;

PROCEDURE SetReadInterleave2;
VAR i: CARDINAL;
BEGIN
    FOR i:=0 TO SectorsPerTrack DIV 2 - 1 DO
        ReadInterleave[i]:=InterleaveUser[i*2];
        ReadInterleave[i+SectorsPerTrack DIV 2]:=InterleaveUser[i*2+1];
    END;
END SetReadInterleave2;

PROCEDURE SetReadInterleave3;
VAR i: CARDINAL;
BEGIN
    FOR i:=0 TO SectorsPerTrack DIV 3 - 1 DO
        ReadInterleave[i]:=InterleaveUser[i*3];
        ReadInterleave[i+SectorsPerTrack DIV 3]:=InterleaveUser[i*3+1];
        ReadInterleave[i+(SectorsPerTrack DIV 3)*2]:=InterleaveUser[i*3+2];
    END;
END SetReadInterleave3;

PROCEDURE SetInterleaveAtari;
VAR i:CARDINAL;
BEGIN
    FormatInterleaveAtari:=TRUE;
    FOR i:=0 TO 17 DO
        InterleaveUser[i]:=InterleaveDD[i];
    END;
    SetReadInterleave2;
END SetInterleaveAtari;

PROCEDURE SetInterleaveED;
VAR i:CARDINAL;
BEGIN
    FormatInterleaveAtari:=TRUE;
    FOR i:=0 TO 25 DO
        InterleaveUser[i]:=InterleaveED[i];
    END;
    SetReadInterleave2;
END SetInterleaveED;

PROCEDURE SetInterleaveSD;
VAR i:CARDINAL;
BEGIN
    FormatInterleaveAtari:=TRUE;
    FOR i:=0 TO 17 DO
        InterleaveUser[i]:=InterleaveSD[i];
    END;
    SetReadInterleave2;
END SetInterleaveSD;

PROCEDURE SetInterleave1;
VAR i:CARDINAL;
BEGIN
    FormatInterleaveAtari:=FALSE;
    FOR i:=0 TO SectorsPerTrack-1 DO
        InterleaveUser[i]:=SHORTCARD(i+1);
    END;
    SetReadInterleave2;
END SetInterleave1;

PROCEDURE SetInterleaveHD;
VAR i:CARDINAL;
BEGIN
    SetInterleaveUser(13);
END SetInterleaveHD;

PROCEDURE SetInterleaveUser(int:CARDINAL);
VAR i,pos:CARDINAL;
BEGIN
    FOR i:=0 TO SectorsPerTrack-1 DO
        InterleaveUser[i]:=255
    END;

    pos:=0;
    FOR i:=1 TO SectorsPerTrack DO
        WHILE InterleaveUser[pos]#255 DO
              pos:=(pos+1) MOD SectorsPerTrack;
        END;
        InterleaveUser[pos]:=SHORTCARD(i);
        pos := (pos+int) MOD SectorsPerTrack;
    END;
    IF (SectorsPerTrack # 36) THEN
       SetReadInterleave2;
    ELSE
       SetReadInterleave3;
    END;
    FormatInterleaveAtari:=TRUE;
END SetInterleaveUser;

PROCEDURE ClearDirDD;
VAR sec:ARRAY[0..255] OF BYTE;
	i:CARDINAL;
BEGIN
	Lib.Fill(ADR(sec),256,0);
	sec[0]:=2;
	sec[1]:=195;
	sec[2]:=2;
	sec[3]:=195;
	sec[4]:=2;
	sec[10]:=15;
	FOR i:=11 TO 54 DO
        sec[i]:=255;
	END;
	sec[56]:=127;
	FOR i:=57 TO 99 DO
	    sec[i]:=255;
	END;
    WriteSector(360,sec);
	Lib.Fill(ADR(sec),256,0);
    i:=361;
    WHILE ok & (i<=368) DO
        WriteSector(i,sec);
        INC(i);
    END;	
END ClearDirDD;

PROCEDURE PrintDebugInfo();
VAR i:CARDINAL;
    value:SHORTCARD;
BEGIN
    IO.WrStr("disk controller configuration:"); IO.WrLn;
    IO.WrStr("   drive = "); IO.WrChar('A'+CHAR(drive)); IO.WrLn;
    IO.WrStr("   BIOS drive_type = "); IO.WrShtCard(drive_type,0); IO.WrLn;
    IO.WrStr("   invert data = "); IO.WrBool(Invert,0); IO.WrLn;
    IO.WrStr("   bytes per sector = "); IO.WrCard(SectorLength,0); IO.WrLn;
    IO.WrStr("   sectors per track = "); IO.WrCard(SectorsPerTrack,0); IO.WrLn;
    IO.WrStr("   tracks per side = "); IO.WrCard(TracksPerSide,0); IO.WrLn;
    IO.WrStr("   sides = "); IO.WrCard(Sides,0); IO.WrLn;
    IO.WrStr("   sector mapping: ");
    CASE CurrentMapping OF
    | Mapping_Track: IO.WrStr("track first");
    | Mapping_Head: IO.WrStr("head first");
    | Mapping_XF551: IO.WrStr("XF551");
    END;
    IO.WrLn;
    IO.WrStr("   drive parameter table:"); IO.WrLn;
    IO.WrStr("   ");
    FOR i:=0 TO 10 DO
        value:=SHORTCARD(new_table[i]);
        IF value < 16 THEN
           IO.WrStr("0");
        END;
        IO.WrShtHex(value,0);
        IO.WrStr(" ");
    END;
    IO.WrLn;
END PrintDebugInfo;


VAR OldCont:PROC;

PROCEDURE Terminate;
BEGIN
    RestoreOldTable;
    OldCont;
END Terminate;

PROCEDURE InitAtariHD;
BEGIN
    DebugLevel := DebugLevel_Quiet;
    Interactive := FALSE;
	SetDriveA;
    SaveOldTable;
    SetNewTable;
    InitDD;
    Lib.Terminate(Terminate,OldCont);
END InitAtariHD;

PROCEDURE TestMapping();
VAR i, maxSector, tmp:CARDINAL;
    s,t,h: SHORTCARD;
BEGIN
    maxSector := SectorsPerTrack * TracksPerSide * Sides;
    FOR i:= 1 TO maxSector DO
        IO.WrCard(i,4);
        IO.WrStr(" -> ");
        SecToSTH(i, s,t,h);
        IO.WrShtCard(s,2); IO.WrChar('/');
        IO.WrShtCard(t,2); IO.WrChar('/');
        IO.WrShtCard(h,1);
        IO.WrStr(" -> ");
        tmp := STHToSec(s,t,h);
        IO.WrCard(tmp,4);
        IF (i = tmp) THEN
           IO.WrStr("  OK");
           IO.WrLn;
        ELSE
           IO.WrStr("  ERROR!");
           IO.WrLn;
           HALT;
        END;
    END;
    IO.WrStr("Mapping seems to be working correctly"); IO.WrLn;
END TestMapping;

END AtariHD.
